/**
 * A robot's view of the board.
 */
public class Board {

    // types of landscape items and drawing them
    public static int EMPTY=0;
    public static int WATER=1;
    public static int WALL=2;
    public static int BASE=3;
    
    public static int NUM_LANDSCAPE_ITEM_TYPES=4;
    
    // used only for drawing landscape items -- these are not used as types 
    // in the GridGraph.
    public static int OTHER=4;
    public static int SELF=5;
    public static int BASEWITHOTHER=6;
    public static int BASEWITHSELF=7;
    public static int PACKAGEDEST=8;
    
    public static char[] GRAPHIC=new char[] {'.','~','#','@','o','s','O','S','x'};
    
    private int m_width;
    private int m_height;
    private GridGraph m_gridGraph;

    /**
     * Initializes the board with the given rows, consisting of '.' for EMPTY,
     * '~' for WATER, '#' for WALL, and '@' for BASE.  Initial weights for
     * WATER and WALL cells are infinite, and EMPTY and BASE are set to 1.
     */
    public Board(String[] rows) {
        m_width=rows[0].length();
        m_height=rows.length;
        m_gridGraph=new GridGraph(m_width, m_height, EMPTY, NUM_LANDSCAPE_ITEM_TYPES);
        for (int y=0; y<m_height; y++ ) {
            for (int x=0; x<m_width; x++) {
                m_gridGraph.setNode(x,y,getType(rows[y].charAt(x)));
            }
        }
        setLandscapeWeight(EMPTY, 1);
        setLandscapeWeight(WATER, -1);
        setLandscapeWeight(WALL, -1);
        setLandscapeWeight(BASE, 1);
    }
    
    public int[] getBaseLocations() {
        // get count first
        int baseCount=0;
        for (int x=1; x<=getWidth(); x++)
            for (int y=1; y<=getHeight(); y++)
                if (getLandscapeAt(x,y)==BASE)
                    baseCount++;
                    
        int[] ret=new int[2*baseCount];
        int i=0;
        for (int x=1; x<=getWidth(); x++) {
            for (int y=1; y<=getHeight(); y++) {
                if (getLandscapeAt(x,y)==BASE) {
                    ret[i++]=x;
                    ret[i++]=y;
                    System.out.println("there is a base at " + x + " " + y);
                }
            }
        }
        return ret;
    }
    
    public int getWidth() {
        return m_width;
    }
    
    public int getHeight() {
        return m_height;
    }


    /**
     * Sets the weight for a given type of landscape item.
     *
     * As a client of this class, it only makes sense to use this method on 
     * EMPTY and BASE, as WALL and WATER's defaults (infinite aka 
     * non-traversable) are the only sensible ones for those types.
     *
     * @param type The landscape item type.
     * @param weight The weight for that type.
     */
    public void setLandscapeWeight(int type, int weight) {
        m_gridGraph.setWeight(type, weight);
    }

    /**
     * Gets the landscape type at a given position.
     *
     * @param x The horizontal position.
     * @param y The vertical position.
     * @return int The type.
     */
    public int getLandscapeAt(int x, int y) {
        return m_gridGraph.getGrid()[x-1][y-1];
    }

    /**
     * Gets the landscape item type indicated by the given character,
     * according to the static GRAPHIC array.
     *
     * @param c The character.
     * @return The landscape item type (WALL if the character is unrecognized)
     */
    private int getType(char c) {
        for (int i=0; i<GRAPHIC.length; i++) {
            if (GRAPHIC[i]==c) {
                return i;
            }
        }
        return WALL;  // shouldn't happen, but if it does we recover by
                      // assuming it's a wall
    }
    
    public void setRobotPosition(int robotId, int x, int y) {
    }

    /**
     * Sets the weight adjustments for the cells near a given landscape item
     * (by type).
     *
     * @param type The type of landscape item to adjust.
     * @param a1 Adjustment for the robot's directly adjacent cells.
     * @param a2 Adjustment for the robot's second level of adjacent cells.
     * @param a3 Adjustment for the robot's third level of adjacent cells.
     */
    public void setLandscapeAdjustment(int type, int a1, int a2, int a3) {
    }
    
    /**
     * Sets the weight adjustments for the cells near a given robot.
     *
     * Adjustments to robots that don't exist on the board have no effect.
     *
     * @param robotId The unique id of the robot.
     * @param o Adjustment for the robot's occupied cell.
     * @param a1 Adjustment for the robot's directly adjacent cells.
     * @param a2 Adjustment for the robot's second level of adjacent cells.
     * @param a3 Adjustment for the robot's third level of adjacent cells.
     */
    public void setRobotAdjustment(int robotId, int o, int a1, int a2, int a3) {
    }
    
    private int getWeight(int x, int y) {
        return m_gridGraph.getWeights()[getLandscapeAt(x,y)];
    }

    public String[] getView(boolean weighted) {
        String[] ret=new String[m_height];
        for (int y=0; y<m_height; y++) {
            StringBuffer row=new StringBuffer(m_width);
            for (int x=0; x<m_width; x++) {
                if (weighted) {
                    int w=getWeight(x+1, y+1);
                    char wChar;
                    if (w<0) { 
                        wChar='*'; 
                    } else {
                        wChar=Character.forDigit(w, 36);
                    }
                    row.append(wChar);
                } else {
                    byte[][] g=m_gridGraph.getGrid();
                    byte tp=g[x][y];
                    row.append(GRAPHIC[(int) tp]);
                }
            }
            ret[y]=row.toString();
        }
        return ret;
    }

    public void printView(boolean weighted) {
        String[] v=getView(weighted);
        for (int i=0; i<v.length; i++) {
            System.out.println(v[i]);
        }
    }
    
    public String[] getPathView(int x1, int y1, int x2, int y2) {
        int[] path=m_gridGraph.getBestPath(x1-1,y1-1,x2-1,y2-1);
        String[] ret=new String[m_height];
        for (int y=0; y<m_height; y++) {
            StringBuffer row=new StringBuffer(m_width);
            for (int x=0; x<m_width; x++) {
                byte[][] g=m_gridGraph.getGrid();
                byte tp=g[x][y];
                if (path!=null) {
                    if (onPath(x,y,path)) {
                        row.append('+');
                    } else {
                        row.append(GRAPHIC[(int) tp]);
                    }
                } else {
                    row.append(GRAPHIC[(int) tp]);
                }
            }
            ret[y]=row.toString();
        }
        return ret;
    }
    
    public boolean onPath(int x, int y, int[] path) {
        for (int i=0; i<path.length; i+=2) {
            if ((path[i]==x) && (path[i+1]==y) ) {
                return true;
            }
        }
        return false;
    }

    public void printPathView(int x1, int y1, int x2, int y2) {
        String[] v=getPathView(x1,y1,x2,y2);
        for (int i=0; i<v.length; i++) {
            System.out.println(v[i]);
        }
    }
    
    public GridGraph getGridGraph() {
        return m_gridGraph;
    }

    public static void main(String[] args) {
        String[] sampleBoard=new String [] {
".....................................................###########...............",
"...............................................................................",
"..........................................................@....................",
"..................~~~~~~~...........#................###########...............",
"..................~~~~~~~...........#........~~~~~~~~~~~~~.....................",
"......................~~~...........#.......~~~~~~~~~~~~~~.....................",
"....................................#..............~~~~~~......................",
"....~~~~~~~~~~~.....................#..............~~~~~~......................",
"....~~~~~~~~~~~.@...................#..........................................",
"....~~~~~~~~~~~.....................#..........................................",
"........................#.#.........######################################.....",
"........................#.#.................................#..................",
"........................#.#.................................#..................",
"#########################.################################.###.################",
"........................#.#..................~~~~~~......#.....#...............",
".....~~~~~~~............#.#................~~~~~~~~~~....#.....#...............",
"...~~~~~~~...............................#~~~~~~~~~......###.###.....~~~~......",
".........................................#~~~~~~~...................~~~~~~~....",
".........................................#~~~~~~~.............######.~~~~~.....",
".........................................#########...........#........~~~~.....",
"........................................................##...#.........~~~.....",
"........................................................#.##.#.................",
"........................................................#...#..................",
"..............................................................................."
};

        Board b=new Board(sampleBoard);
        
        
     //   b.setType(3,3,Board.SELF);
     //   b.setType(64,20,Board.PACKAGEDEST);
        b.printView(true);
        System.out.println("");
        b.printPathView(3,3,64,20);
    }

}
