import java.util.Stack;

/**
 * A special type of grid that can be treated as a graph.
 * <p></p>
 * It's a very simple graph.  Arcs go up, down, left, and right.
 * Arc weights are derived by node weights.  The weight of the arc 
 * from point A to B is weight(A)+weight(B).
 * <pre>
 * (0,0)
 *      0--[2]--0--[2]--0--[2]--0
 *      |       |       |       |
 *     [2]     [2]     [2]     [2]
 *      |       |       |       |
 *      0--[2]--0--[2]--0--[2]--0
 *      |       |       |       |
 *     [2]     [2]     [2]     [2]
 *      |       |       |       |
 *      0--[2]--0--[2]--0--[2]--0
 *                                (3,2)                      
 * </pre>
 * In the above graph, dashes are arcs, numbers in [brackets] are
 * arc weights, and lone numbers are nodes (aka cells).  The number
 * of the cell indicates it's type.  Each type of node has a weight.
 * In the above example, nodes of type "0" have a weight of 1.
 * Therefore, all arcs from type "0" nodes to type "0" nodes have
 * a weight of 2, as indicated by the numbers in brackets.
 * <p></p>
 * Let's look at the same sized grid with two other node types.
 * We will let nodes of type "1" have a weight of two, and nodes
 * of type "2" will have a weight of Infinity (represented below as "I" --
 * but in this class, the value "infinity" is actually represented by
 * any negative number).
 * <p></p>
 * <pre>
 * (0,0)
 *      0--[3]--1--[4]--1--[3]--0
 *      |       |       |       |
 *     [2]     [I]     [I]     [2]
 *      |       |       |       |
 *      0--[I]--2--[I]--2--[I]--0
 *      |       |       |       |
 *     [2]     [I]     [I]     [2]
 *      |       |       |       |
 *      0--[2]--0--[3]--1--[3]--0
 *                                (3,2)                      
 * </pre>
 * Note that since arcs of infinite weight can never be traversed,
 * the graph can be simplified as:
 * <pre>
 * (0,0)
 *      0--[3]--1--[4]--1--[3]--0
 *      |                       |
 *     [2]                     [2]
 *      |                       |
 *      0                       0
 *      |                       |
 *     [2]                     [2]
 *      |                       |
 *      0--[2]--0--[3]--1--[3]--0
 *                                (3,2)
 * </pre>
 * The entire reason for this class' existence is the public instance 
 * method, <code>getBestPath(...)</code>.  It uses an algorithm called
 * Dijstra's Algorithm to find the best (least weighted path) between
 * two points in the grid.
 * <p></p>
 * Let's use the above grid as an example, and say we want to go
 * from point 0,0 to 3,2.  So accordingly, we'd call:
 * <pre>
 * int[] path=graph.getBestPath(0,0,3,2);
 * </pre> 
 * This would return the integer array:
 * <pre>
 * {0,1, 0,2, 1,2, 2,2, 3, 2}
 * </pre>
 * .. because with this path, we incur a cost of 12, whereas
 * the other path would cost 14.
 * <p></p>
 * This is a pretty simple example -- with only two reasonable
 * paths, you can just follow them both, add up the values, and
 * compare.  Dijstra's algorithm comes in handy when looking at
 * much more complicated graphs, where there are many more possible
 * paths.
 *
 * @author cwilper@cs.cornell.edu
 */                     
public class GridGraph {

    private static final int UNDEFINED=-1;
    private static final int INFINITY=-2;

    private int m_xSize;
    private int m_ySize;
    private int m_lastStartX;
    private int m_lastStartY;
    private byte[][] m_grid;
    private int[] m_weights;

    private int[][] m_bestPathWeightFinished;
    private int[][] m_bestPathWeightUnfinished;

    private int[][] m_bestBuddyX;
    private int[][] m_bestBuddyY;
    
    /**
     * Creates a new GridGraph, initializing each node as being type fillNodeType.
     *
     * @param xSize The size of a grid (left-right)
     * @param ySize The size of the grid (down-up)
     */
    public GridGraph(int xSize, int ySize, int fillNodeType, int numTypes) {
        m_weights=new int[numTypes];
        m_grid=new byte[xSize][ySize];
        for (int x=0; x<xSize; x++) {
            for (int y=0; y<ySize; y++) {
                m_grid[x][y]=(byte) fillNodeType;
            }
        }
        m_xSize=xSize;
        m_ySize=ySize;
        m_bestPathWeightUnfinished=new int[xSize][ySize];
        m_bestPathWeightFinished=new int[xSize][ySize];
        m_bestBuddyX=new int[xSize][ySize];
        m_bestBuddyY=new int[xSize][ySize];
    }
    
    public byte[][] getGrid() {
        return m_grid;
    }
    
    public int[] getWeights() {
        return m_weights;
    }

    /**
     * Sets the weight of a certain type of node.
     *
     * The weight of a node is the weight of all edges
     * leading to and from it in the grid.
     * 
     * @param nodeType The identifier for the type. 
     * @param weight The weight, less than zero meaning infinity.
     */
    public void setWeight(int nodeType, int weight) {
        m_weights[nodeType]=weight;
    }

    /**
     * Sets a point in the grid with a node of the given type.
     *
     * @param xPos The x position, starting with 0.
     * @param yPos The y position, starting with 0.
     */
    public void setNode(int xPos, int yPos, int type) {
        m_grid[xPos][yPos]=(byte) type;
    }

    /**
     * Convenience method for <code>getBestPath</code> with no adjustments.
     *
     * @param startX The starting x coordinate.
     * @param startY The starting y coordinate.
     * @param finishX The starting x coordinate.
     * @param finishY The starting y coordinate.
     */
    public int[] getBestPath(int startX, int startY, int finishX, int finishY) {
        return getBestPath(startX, startY, finishX, finishY, null);
    }

    /**
     * Gets the best path from one node to another on the graph, using
     * Dijkstra's algorithm.  
     * <p></p>
     * Results are returned as an even-sized array of integers which 
     * represent subsequent points in the path, beginning with the point 
     * immediately adjacent to the starting point and ending with the 
     * finishing point.  
     * <p></p>
     * <ul>
     *   <li> If the starting point <i>is</i> the finishing point, an 
     *        empty array is returned.</li>
     *   <li> If there is no path, null is returned.
     * </ul>
     *
     * @param startX The starting x coordinate.
     * @param startY The starting y coordinate.
     * @param finishX The starting x coordinate.
     * @param finishY The starting y coordinate.
     * @param adj Weight adjustments to apply while computing node weights.
     */  
    public int[] getBestPath(int startX, int startY, int finishX, int finishY, int[][] adj) {
        dijkstra(startX, startY, finishX, finishY, adj);
        m_lastStartX=startX;
        m_lastStartY=startY;
        // traverse from finish to start, pushing on stack,
        // exiting if any -1 (x or y coords) for bestBuddy exist (means no path)
        int thisX=finishX;
        int thisY=finishY;
        Stack s=new Stack();
        int oldThisX;
        s.push(new Integer(thisY));
        s.push(new Integer(thisX));
        while (!((thisX==startX) && (thisY==startY))) {
            oldThisX=thisX;
            thisX=m_bestBuddyX[oldThisX][thisY];
            thisY=m_bestBuddyY[oldThisX][thisY];
            if (thisX<0 || thisY<0) {
                return null;
            }
            if ( !((thisX==startX) && (thisY==startY)) ) {
            s.push(new Integer(thisY));
            s.push(new Integer(thisX));
            }
        }
        
        int[] ret=new int[s.size()];
        int c=0;
        while (!s.isEmpty()) {
            ret[c++]=((Integer) s.pop()).intValue();
        }
        return ret;
    }

    /**
     * Get the cost going from the specified points in the previously
     * computed path.
     * <p></p>
     * This must be called after making a call to getBestPath() and verifying
     * that the result is non-null, and knowing that the path from
     * startX,startY to finishX,finishY exists in the most recently computed
     * best path.  It's important to use the same order provided in the
     * path.
     *
     * @param finishX The ending point x coordinate.
     * @param finishY The ending point y coordinate.
     * @param startX The starting point x coordinate.
     * @param startY The starting point y coordinate.
     */
    public int getCost(int finishX, int finishY, int startX, int startY) {
        if (startX==m_lastStartX && startY==m_lastStartY) {
            return m_bestPathWeightFinished[finishX][finishY];
        }
        return m_bestPathWeightFinished[finishX][finishY]-
                m_bestPathWeightFinished[startX][startY];
    }

    /**
     * Perform dijstra's algorithm to find the least expensive path between
     * two nodes in the graph.
     * <p></p>
     * When finished, the path can be obtained by following the m_bestBuddyX[]
     * and m_bestBuddyY[] coordinates, starting at the finish and ending at
     * the start.  If there is no possible path (one of the nodes has a weight
     * of INFINITY [represented as anything below 0]), one of the bestBuddy
     * coordinates in the path from finish to start will be &lt;0.
     */
    private void dijkstra(int startX, int startY, int finishX, int finishY, int[][] adj) {
        // init, all=-1 except m_best(startNode)=0
        int x;
        int y;
        for (x=0; x<m_xSize; x++) {
            for (y=0; y<m_ySize; y++) {
                m_bestBuddyX[x][y]=UNDEFINED;
                m_bestBuddyY[x][y]=UNDEFINED;
                m_bestPathWeightUnfinished[x][y]=INFINITY;
                m_bestPathWeightFinished[x][y]=INFINITY;
            }
        }
        m_bestPathWeightUnfinished[startX][startY]=0;
        boolean finishedFinish=false;
        while (true) {
            // compute the minimum estimated weight from unfinished
            //  - if all UNDEFINED, return (done)
            int bestUnfinishedX=-1;
            int bestUnfinishedY=-1;
            int bestUnfinished=INFINITY;
            int candidate;
            for (x=0; x<m_xSize; x++) {
                for (y=0; y<m_ySize; y++) {
                    candidate=m_bestPathWeightUnfinished[x][y];
                    if ( candidate >= 0)  {  // not undefd, not inf
                        if ( (bestUnfinished==INFINITY) || (candidate<bestUnfinished) ) {
                            bestUnfinishedX=x;
                            bestUnfinishedY=y;
                            bestUnfinished=candidate;
                        }
                    }
                }
            }
            if (bestUnfinishedX==-1) {
                // no more in unfinished
                return;
            }
            // we found one... add to finished and remove from unfinished
            m_bestPathWeightFinished[bestUnfinishedX][bestUnfinishedY]=bestUnfinished;
            // but before bothering with the next step, if it's finishX, finishY node, just return
            if ( (bestUnfinishedX==finishX) && (bestUnfinishedY==finishY) ) {
                return;
            }
            // ok, now continue.  remove from unfinished.
            m_bestPathWeightUnfinished[bestUnfinishedX][bestUnfinishedY]=UNDEFINED;
            // relax unfinished neighbors.. there's a max of 4 (up down left right)
            int nx, ny, p, o;
            // up
            nx=bestUnfinishedX;
            ny=bestUnfinishedY-1;
            int buNodeWeight=m_weights[m_grid[bestUnfinishedX][bestUnfinishedY]];
            int nWeight;
            if ((ny>=0)) { 
                nWeight=m_weights[m_grid[nx][ny]];
                if (nWeight>-1) {
                p=bestUnfinished + buNodeWeight + nWeight;
                o=m_bestPathWeightUnfinished[nx][ny];
                if (( o==INFINITY) || (p < o)) {
                m_bestPathWeightUnfinished[nx][ny]=p;
                m_bestBuddyX[nx][ny]=bestUnfinishedX;
                m_bestBuddyY[nx][ny]=bestUnfinishedY;
                }
                }
            }
            // down
            nx=bestUnfinishedX;
            ny=bestUnfinishedY+1;
            if ((ny<m_ySize)) {
                nWeight=m_weights[m_grid[nx][ny]];
                if (nWeight>-1) {
                p=bestUnfinished + buNodeWeight + nWeight;
                o=m_bestPathWeightUnfinished[nx][ny];
                if (( o==INFINITY) || (p < o)) {
                m_bestPathWeightUnfinished[nx][ny]=p;
                m_bestBuddyX[nx][ny]=bestUnfinishedX;
                m_bestBuddyY[nx][ny]=bestUnfinishedY;
                }
                }
            }
            // left
            nx=bestUnfinishedX-1;
            ny=bestUnfinishedY;
            if ((nx>=0)) {
                nWeight=m_weights[m_grid[nx][ny]];
                if (nWeight>-1) {
                p=bestUnfinished + buNodeWeight + nWeight;
                o=m_bestPathWeightUnfinished[nx][ny];
                if (( o==INFINITY) || (p < o)) {
                m_bestPathWeightUnfinished[nx][ny]=p;
                m_bestBuddyX[nx][ny]=bestUnfinishedX;
                m_bestBuddyY[nx][ny]=bestUnfinishedY;
                }
                }
            }
            // right
            nx=bestUnfinishedX+1;
            ny=bestUnfinishedY;
            if ((nx<m_xSize)) {
                nWeight=m_weights[m_grid[nx][ny]];
                if (nWeight>-1) {
                p=bestUnfinished + buNodeWeight + nWeight;
                o=m_bestPathWeightUnfinished[nx][ny];
                if (( o==INFINITY) || (p < o)) {
                m_bestPathWeightUnfinished[nx][ny]=p;
                m_bestBuddyX[nx][ny]=bestUnfinishedX;
                m_bestBuddyY[nx][ny]=bestUnfinishedY;
                }
                }
            } 
        }
    }
    
    public void printPath(int x1, int y1, int x2, int y2) {
        long s=System.currentTimeMillis();
        int[] b=getBestPath(x1,y1,x2,y2);        
        long t=System.currentTimeMillis()-s;
        System.out.println("In " + t + " milliseconds, I determined the path from " + x1 + "," + y1 + " to " + x2 + "," + y2);
        if (b==null) {
            System.out.println("It's impossible!");
        } else {
            for (int i=0; i<b.length; i+=2) {
                System.out.print("->(" + b[i] + "," + b[i+1] + ")");
            }
            System.out.println(" (total cost=" + getCost(x2, y2, x1, y1) + ")");
        }
    }

    /**
     * Does a basic test of this class, showing how it might be used.
     *
     * @param args Command-line parameters (unused)
     */
    public static void main(String[] args) {
        GridGraph g=new GridGraph(3,3,1,5);
        g.setWeight(1,1);
        g.setWeight(2,2);
        g.setWeight(3,-1);

        /*
         212
         2I1   Build a graph with the indicated weights (I=INFINITY)
         I21
        */

        g.setNode(0,0,2);
        g.setNode(0,1,2);
        g.setNode(0,2,3);
        g.setNode(1,0,1);
        g.setNode(1,1,3);
        g.setNode(1,2,2);
        g.setNode(2,0,2);
        g.setNode(2,1,1);
        g.setNode(2,2,1);

        int startX=1;
        int startY=2;
        int finishX=0;
        int finishY=1;
        long s=System.currentTimeMillis();
        int[] b=g.getBestPath(1,2,0,1);        
        long t=System.currentTimeMillis()-s;
        System.out.println("In " + t + " milliseconds, I determined:");
        if (b==null) {
            System.out.println("It's impossible!");
        } else {
            for (int i=0; i<b.length; i+=2) {
                System.out.print("->(" + b[i] + "," + b[i+1] + ")");
            }
            System.out.println(" (total cost=" + g.getCost(finishX, finishY, startX, startY) + ")");
        }
        
    }

}